
structure RobUtil =
struct
    open World
    
    fun getpackage packages pid = valOf (World.Map.find (packages, pid))
    fun getbot robots rid = valOf (World.Map.find (robots, rid))

    (* how much weight is a robot carrying (at a minimum) *)
    fun carried (board as BOARD{packages,...}) (ROBOT {packages=ps, ...}) =
	foldl (fn (PACKAGE{weight = Known w, ...}, tot) => w + tot
	         | (_, tot) => tot) 0 (map (getpackage packages) ps)

    (* find a spot that satisfies a predicate *)
    exception Found of int * int
    fun findspot (board as BOARD{layout=(b, width, height),...}) f =
	let in
	    Util.for 1 (width-2)
	    (fn c =>
	     Util.for 1 (height-2)
	     (fn r => if f (r, c, sub (b, width, height) (r, c)) then raise Found (r, c)
	      else ()));
	    NONE
	end handle Found loc => SOME loc

    (* find a list of spots that satisfy a predicate *)
    fun findspots (board as BOARD{layout=(b, width, height),...}) f =
	let
	    fun findspots' 0 _ l = l
              | findspots' r 0 l = findspots' (r-1) (width-2) l
              | findspots' r c l = 
		let
                  val l' = if f (r, c, sub (b, width, height) (r, c)) then ((r,c)::l) else l 
                in
                  findspots' r (c-1) l'
                end
        in
	    findspots' (height-2) (width-2) nil
	end handle _ => [(1,1)] (* bogus: protection *)

    (* package ids that belong here *)
    fun packsdesthere loc packages rp =
	(List.filter (fn PACKAGE{dest = Known dest, ...} => dest = loc | _ => false) (map (getpackage packages) rp))
	handle _ => nil
	
    (* packages that are on the ground right here *)
    fun packsarehere loc packages =
	(List.filter (fn PACKAGE{loc = Known (Ground here), ...} => here = loc | _ => false) 
	     (World.Map.listItems packages)) handle _ => nil

end
