
structure StupidBotII :> ROBOT =
struct
    open Comm
    open World

    datatype plan = 
	GOTO of int * int

    type mind = plan

    (* find a base to GOTO (except for the one listed) *)
    fun goto_base_not (a,b) (board as BOARD{layout, cap, ...}) (me as ROBOT{loc, ...}) =
	case RobUtil.findspot board (fn (r, c, SPOT{space=Dot true,...}) => (r <> a orelse b <> c) | _ => false) of
	    SOME l => GOTO l
	  | NONE => 
		(case RobUtil.findspot board (fn (r, c, SPOT{space=Dot _,...}) => (r <> a orelse b <> c) | _ => false) of
		   SOME l => GOTO l
                 | NONE => GOTO (1,1)) (* ??? *)
					     

    val init = goto_base_not (~1, ~1)
    val itos = Int.toString
    fun dtos (a,b) = itos a ^ "," ^ itos b
    fun pltos (Ground d) = "@" ^ dtos d
      | pltos (Robot r) = "[rob " ^ itos r ^ "]"

    fun ptos (PACKAGE{id,loc,weight,dest}) = itos id

    fun ktos f (Known n) = f n
      | ktos _ _ = "Unknown"

    fun move (board as BOARD{layout, packages, robots, cap}) dmap (GOTO(a, b)) (me as ROBOT{loc as (r, c), packages=packs, ...}) =
	let in
	    print ("(StupidBot 2.0) I'm at " ^ itos r ^ "," ^ itos c ^ " heading to " ^ itos a ^ "," ^ itos b ^ "\n");
	    print ("Location of known packages:\n     " ^
		   StringUtil.delimit "\n     " (map (fn PACKAGE{id, loc, weight, dest} =>
						      itos id ^ ": " ^ ktos dtos dest ^ " " ^ ktos itos weight ^ "lb. "
						      ^ ktos pltos loc) (World.Map.listItems packages)) ^ "\n");
	if a = r andalso b = c then
	    (* got there! *)
	    (* pick up or put down? *)
	    (case RobUtil.packsdesthere loc packages packs of
		nil => (case packs of
                          nil => (* empty handed!! find a new base? *)
                            (case RobUtil.packsarehere loc packages of
                               nil => 			
                                 let in
                                   print ("Why am I here? Finding a new base...\n");
                                   move board dmap (goto_base_not (a,b) board me) me
                                 end
                             | l => 
                                 let in
                                   print ("Pick up: " ^ StringUtil.delimit " " (map ptos l) ^ "\n");
                                   (GOTO(a,b), 1, Pick l)
                                 end)
                        | (pid::_) => 
                               let val PACKAGE{dest = Known dest, ...} = RobUtil.getpackage packages pid
                               in 
                                 print ("Now proceeding to " ^ dtos dest ^ " ...\n");
                                 move board dmap (GOTO dest) me
                               end)
	      | l => 
		    let in
			print ("Deliver: " ^ StringUtil.delimit " " (map ptos l) ^ "\n");
			(GOTO(a,b), 1, Drop l)
		    end)
	else
	    let 
		fun eval ((dr, dc), m) =
		    case sub layout (r + dr, c + dc) of
		        SPOT {space = Dot _, robots = ro} => (* closer to goto? *)
			    ((if dr > 0 andalso a > r then 100
			      else if dr < 0 andalso a < r then 100 
				   else 0) +
			     (if dc > 0 andalso b > c then 100
			      else if dc < 0 andalso b < c then 100
				   else 0), m)
		      | SPOT {space = Water, ...} => (~1000, m)
		      | SPOT {space = Wall, ...} => (~20, m)
			    
		val moves = map eval [((1,0),  N),
				      ((~1,0), S),
				      ((0,1),  E),
				      ((0,~1), W)]
		    
		val l = rev (ListUtil.sort (ListUtil.byfirst Int.compare) moves)

		fun mtos S = "S"
		  | mtos E = "E"
		  | mtos W = "W"
		  | mtos N = "N"

		val _ = app (fn (v, m) =>
			     print ("   " ^ mtos m ^ ": " ^ itos v ^ "\n")) l
	    in
		(GOTO(a,b), 1, Move (#2 (hd l)))
	    end
	end
end
